@@ -144,9 +144,12 @@ Rboolean Rcairo_new_device_driver(NewDevDesc *dd, const char *type, int conn, co
 #endif
 
 #if R_GE_version >= 13
-    dd->deviceVersion = R_GE_definitions;
+	dd->deviceVersion = R_GE_definitions;
+#endif
+#if R_GE_version >= 16
+	/* FIXME: we don't actually implement 14 and 15 */
+	dd->deviceVersion = R_GE_glyphs;
 #endif
-
     dd->deviceSpecific = (void *) xd;
 
 #ifdef JGD_DEBUG
@@ -273,11 +276,17 @@ SEXP cairo_create_new_device(SEXP args)
 	    free(dev);
 	    error("unable to start device %s", devname);
 	}
-	
+
+	dd = GEcreateDevDesc(dev);
+#if R_VERSION >= R_Version(3,0,0)
+	/* technically, this was added in R 2.7.0 (undocumented), but it is really
+	   needed only since 4.4.0 when API cleanup removed gsetVar() */
+	GEaddDevice2(dd, devname);
+#else
+	GEaddDevice(dd);
 	gsetVar(PROTECT(install(".Device")), PROTECT(mkString(devname)), R_NilValue);
 	UNPROTECT(2);
-	dd = GEcreateDevDesc(dev);
-	GEaddDevice(dd);
+#endif
 	GEinitDisplayList(dd);
     } END_SUSPEND_INTERRUPTS;
 #ifdef JGD_DEBUG
@@ -304,7 +313,7 @@ void gdd_get_version(int *ver) {
 }
 
 SEXP cairo_font_match(SEXP args){
-#if CAIRO_HAS_FT_FONT
+#if USE_CAIRO_FT
 	SEXP v;
 	const char *fcname;
 	int sort;
@@ -418,22 +427,32 @@ SEXP cairo_font_match(SEXP args){
 }
 
 SEXP cairo_font_set(SEXP args){
-#if CAIRO_HAS_FT_FONT
+#if USE_CAIRO_FT
 	SEXP v;
 	int i;
 	const char *font;
 
 	args = CDR(args);
 
-	/* regular font */
+	/* 0-3 = regular, 4 = symbol, 5 = PUA */
 	for (i = 0; i < 5; i++){
 		v = CAR(args); args = CDR(args);
 		if (!isNull(v) && isString(v) && LENGTH(v)==1){
 			font = CHAR(STRING_ELT(v,0));
-			Rcairo_set_font(i,font);
+			Rcairo_set_font(i, font);
 		}
 	}
+	v = CAR(args);
+	if (!isNull(v)) {
+		int pua = asInteger(v) ? 1 : 0;
+#if R_VERSION >= R_Version(4,0,0)
+		Rcairo_symbol_font_use_pua = pua;
 #else
+		if (!pua)
+			warning("You R does not support symbol font without PUA, upgrade R to at least 4.0.0");
+#endif /* !PUA (R 4.0.0) */
+	}
+#else  /* FT_FONT */
 #ifdef WIN32
 	warning("CairoFonts() has no effect on Windows. Please use par(family=\"...\") to specify the desired font - see ?par.");
 #else
