@@ -110,6 +110,16 @@ static void     CairoGD_releaseClipPath(SEXP ref, pDevDesc dd);
 static SEXP     CairoGD_setMask(SEXP path, SEXP ref, pDevDesc dd);
 static void     CairoGD_releaseMask(SEXP ref, pDevDesc dd);
 
+#if R_GE_version >= 16
+/* this is required since 14 otherwise R segfaults */
+static SEXP CairoGD_Capabilities(SEXP capabilities);
+
+/* GE 16 glyph API */
+static void CairoGD_Glyph(int n, int *glyphs, double *x, double *y,
+						  SEXP font, double size,
+						  int col, double rot, pDevDesc dd);
+#endif
+
 /* fake mbcs support for old R versions */
 #if R_GE_version < 4
 #define mbcslocale 0
@@ -315,7 +325,7 @@ static void Rcairo_setup_font(CairoGDDesc* xd, R_GE_gcontext *gc) {
 				strcat(spec, specs[i - 5]);
 				Rcairo_set_font(i, spec);
 			} else
-				Rcairo_set_font(i, spec);
+				Rcairo_set_font(i, gc->fontfamily);
 			if (Rcairo_fonts[i].family)
 				free(Rcairo_fonts[i].family);
 			Rcairo_fonts[i].family = strdup(gc->fontfamily);
@@ -783,7 +793,7 @@ static rc_text_shape *c_setup_glyphs(CairoGDDesc *xd, R_GE_gcontext *gc, const c
 	ulen = str2utf16(str, strlen(str), &text, encoding) / sizeof(UChar); /* str2utf16 returns bytes, need chars */
 	ubidi_setPara(bidi, text, ulen, UBIDI_DEFAULT_LTR, NULL, &err);
 	if (U_FAILURE(err))
-		Rf_error("Unable to compute UBiDi for string '%'", str);
+		Rf_error("Unable to compute UBiDi for string '%s'", str);
 
 	rc_text_shape *ts = init_text_shape();
 	int i = xd->fontface - 1;
@@ -1419,6 +1429,82 @@ static void CairoGD_TextNative(double x, double y, constxt char *str,
 	CairoGD_TextEnc(x, y, str, rot, hadj, gc, dd, "");
 }
 
+#if R_GE_version >= 16
+static SEXP CairoGD_Capabilities(SEXP capabilities) {
+	SEXP glyphs = PROTECT(ScalarInteger(1));
+	SET_VECTOR_ELT(capabilities, R_GE_capability_glyphs, glyphs);
+	UNPROTECT(1);
+	return capabilities;
+}
+
+static void CairoGD_Glyph(int n, int *glyphs, double *x, double *y,
+						  SEXP font, double size,
+						  int col, double rot, pDevDesc dd) {
+	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
+	cairo_t *cc;
+
+	if (!xd || !xd->cb) return;
+
+	cc = xd->cb->cc;
+
+	/* font info is explicit so we don't use our regular cache and mapping */
+	double weight = R_GE_glyphFontWeight(font);
+	int style = R_GE_glyphFontStyle(font);
+	int wt = (weight > 400) ? CAIRO_FONT_WEIGHT_BOLD : CAIRO_FONT_WEIGHT_NORMAL;
+	int sl = (style == R_GE_text_style_normal) ? CAIRO_FONT_SLANT_NORMAL : CAIRO_FONT_SLANT_ITALIC;
+
+#ifdef JGD_DEBUG
+	Rprintf("INFO: Glyph(n=%d), fontfile='%s', index=%d, weight=%f, style=%d, col=%x, rot=%f\n",
+			n, R_GE_glyphFontFile(font) ? R_GE_glyphFontFile(font) : "<NULL>",
+			R_GE_glyphFontIndex(font), weight, style, col, rot);
+#endif
+
+	cairo_font_face_t *cairo_face = NULL;
+#if USE_CAIRO_FT
+	FcPattern *pattern = FcPatternBuild(NULL,
+										FC_FILE, FcTypeString,
+										R_GE_glyphFontFile(font),
+										FC_INDEX, FcTypeInteger,
+										R_GE_glyphFontIndex(font),
+										NULL);
+	cairo_face = cairo_ft_font_face_create_for_pattern(pattern);
+	FcPatternDestroy(pattern);
+#endif
+	if (cairo_face &&
+		cairo_font_face_status(cairo_face) == CAIRO_STATUS_SUCCESS) {
+		cairo_set_font_face(cc, cairo_face);
+	} else /* fall back - hopefully not needed */
+		cairo_select_font_face(cc, R_GE_glyphFontFamily(font), sl, wt);
+
+	/* FIXME: not sure if we need to scale - Quartz uses 72 baseline - check sizes */
+	cairo_set_font_size (cc, size / (72*dd->ipr[0]));
+
+	Rcairo_set_color(cc, col);
+
+	int i = 0;
+	while (i < n) {
+		if (rot != 0.0) {
+			cairo_save(cc);
+			cairo_translate(cc, x[i], y[i]);
+			cairo_rotate(cc, -rot/180.*M_PI);
+			cairo_translate(cc, -x[i], -y[i]);
+		}
+
+		cairo_glyph_t cairoGlyph;
+		cairoGlyph.index = glyphs[i];
+		cairoGlyph.x = x[i];
+		cairoGlyph.y = y[i];
+
+		cairo_show_glyphs(cc, &cairoGlyph, 1);
+		/* for path ops use cairo_glyph_path(cc, &cairoGlyph, 1); instead */
+
+		if (rot != 0.0)
+			cairo_restore(cc);
+		i++;
+	}
+}
+#endif
+
 static SEXP CairoGD_setPattern(SEXP pattern, pDevDesc dd) {
     return R_NilValue;
 }
@@ -1475,6 +1561,11 @@ void Rcairo_setup_gd_functions(NewDevDesc *dd) {
     dd->releaseClipPath = CairoGD_releaseClipPath;
     dd->setMask         = CairoGD_setMask;
     dd->releaseMask     = CairoGD_releaseMask;
+#if R_GE_version >= 16
+	/* NOTE: caps must be set from 14 on or else R segfaults */
+	dd->capabilities    = CairoGD_Capabilities;
+	dd->glyph           = CairoGD_Glyph;
+#endif
 #endif
 #endif
 #endif
